/*
** Module   : NativeSequence.java
** Abstract : Static class that handles sequence functions and statements 
**             from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 OM  20121116          Created initial version with CURRVAL, NEXTVAL 
**                           and SETVAL support only.
*/

package com.goldencode.p2j.persist;

import java.sql.*;
import java.util.*;

import com.goldencode.p2j.persist.dialect.P2JDialect;
import com.goldencode.p2j.util.integer;


/**
 * Implementation of sequences using a SQL table and manually handle sequence
 * primitives using SQL queries. The calls from converted P2J code are 
 * delegated to protected method implementations.
 * 
 * The structure of the table is the following (works both H2 and PostgreSQL):
 * CREATE TABLE _P2JSequences_ (
 *    id integer NOT NULL,
 *    name text NOT NULL, -- name of the sequence
 *    initial bigint NOT NULL, -- initial value
 *    increment bigint NOT NULL, -- the step to generate next value
 *    maxval bigint, -- upper bound
 *    minval bigint, -- lower bound
 *    currval bigint NOT NULL, -- current value
 *    cycle boolean NOT NULL DEFAULT false, -- cycling sequence ?
 *    unknown boolean NOT NULL DEFAULT false, -- has cross bounds ?
 *    CONSTRAINT "_P2JSequences__pkey" PRIMARY KEY (id)
 * );
 */
public class CustomSequence extends Sequence
{
   /**
    * The list of sequences grouped by databases.
    */
   private HashMap<String, LinkedList<String>> sequences = 
         new HashMap<String, LinkedList<String>>();

   /**
    * Initialize the internal structures.
    * To be called when first connect to a database in order to read 
    * the list of current sequences from the table.
    * @param   ldbName
    *          The logical name of the database.
    */
   protected boolean init(String ldbName)
   {
      // make sure the structure is initialized for this database:
      LinkedList<String> crtSeqList = null;
      if (!sequences.containsKey(ldbName)) 
      {
         crtSeqList = new LinkedList<String>();
         sequences.put(ldbName, crtSeqList);
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      // P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = "SELECT name FROM \"_P2JSequences_\";";
      // H2Dialect: "SELECT name FROM _P2JSequences_;"

      try
      {
         ResultSet rs = persistence.executeSQLQuery(strQuery);
         while (rs.next())
         {
            crtSeqList.add(rs.getString(1));
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
      catch (SQLException sqle)
      {
         sqle.printStackTrace();
      }
      return true;
   }

   /**
    * Queries the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has already passed the maximum value.
    */
   protected integer internalCurrentValue(String seqName, String ldbName)
   {
      if (ldbName == null)
      {
         for (String db : sequences.keySet())
         {
            if (sequences.get(db).contains(seqName))
            {
               // I have found the database to which the sequence belongs to
               ldbName = db;
               break;
            }
         }

         // let supper class handle this:
         if (ldbName == null)
         {
            ldbName = getDatabaseForSequence(seqName);
         }

         // if still cannot find the appropriate database, return unknown
         if (ldbName == null)
         {
            return new integer();
         }
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      // P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = 
            "SELECT currval, unknown " +
            "FROM \"_P2JSequences_\" " +
            "WHERE name='" + seqName + "';";
      // H2 Dialect: "SELECT currval, unknown FROM _P2JSequences_ WHERE name='my-seq-1';"
      integer ret = new integer();

      try
      {
         ResultSet rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            boolean isUnknown = rs.getBoolean(2);
            if (!isUnknown)
            {
               // is is Unknown we don't read the actual value any more
               ret = new integer(rs.getLong(1));
            }
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
      catch (SQLException sqle)
      {
         sqle.printStackTrace();
      }

      return ret;
   }

   /**
    * Computes and returns the next value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has passed the maximum/minimum value.
    */
   protected integer internalNextValue(String seqName, String ldbName)
   {
      if (ldbName == null)
      {
         for (String db : sequences.keySet())
         {
            if (sequences.get(db).contains(seqName))
            {
               //I have found the database to which the sequence belongs to
               ldbName = db;
               break;
            }
         }

         // let supper class handle this:
         if (ldbName == null)
         {
            ldbName = getDatabaseForSequence(seqName);
         }

         // if still cannot find the appropriate database, return unknown
         if (ldbName == null)
         {
            return new integer();
         }
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      // P2JDialect dialect = (P2JDialect) persistence.getDialect();
      integer ret = new integer(); // unknown result, (TODO: promote it to int64)

      try
      {
         persistence.beginTransaction();
         // step1: get sequence details:
         ResultSet rs = persistence.executeSQLQuery(
               "SELECT increment, maxval, minval, " + 
                        "currval, cycle, unknown " +
               "FROM \"_P2JSequences_\" " +
               "WHERE name='" + seqName + "';");
         // H2 Dialect: "SELECT increment, maxval, minval, currval, cycle, unknown FROM _P2JSequences_ WHERE name='my-seq-1';"
         if (rs.next())
         {
            long seqInc = rs.getLong(1);
            long seqMax = rs.getLong(2);
            long seqMin = rs.getLong(3);
            long curVal = rs.getLong(4);
            boolean seqCycle = rs.getBoolean(5);
            boolean seqUnknown = rs.getBoolean(6);
            if (!seqUnknown)
            {
               curVal += seqInc;
               if (seqInc > 0)
               {
                  if (curVal > seqMax)
                  {
                     if (seqCycle)
                     {
                        //warp around:
                        curVal = seqMin;
                     }
                     else
                     {
                        // max value passed by
                        seqUnknown = true;
                     }
                  }
               }
               else if (seqInc < 0)
               {
                  if (curVal < seqMin)
                  {
                     if (seqCycle)
                     {
                        // warp around:
                        curVal = seqMax;
                     }
                     else
                     {
                        // min value passed by
                        seqUnknown = true;
                     }
                  }
               }

               // prepare the value to be returned if not computed as unknown
               if (!seqUnknown)
               {
                  ret = new integer(curVal);
               }

               // update into database:
               persistence.executeSQL(
                     "UPDATE \"_P2JSequences_\" " +
                     "SET currval = " + curVal + ", " +
                        "unknown = " + seqUnknown + " " +
                     "WHERE name='" + seqName + "';");
               // H2 Dialect: "UPDATE  _P2JSequences_ SET currval = 5, unknown = false WHERE name='my-seq-1';"
            }
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
      catch (SQLException sqle)
      {
         sqle.printStackTrace();
      }
      finally
      {
         try
         {
            persistence.commit();
         }
         catch (PersistenceException e)
         {
            e.printStackTrace();
         }
      }

      return ret;
   }

   /**
    * Initialize (reset) the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   protected void internalSetValue(String seqName,
                                    String ldbName,
                                    long newVal)
   {
      if (ldbName == null)
      {
         for (String db : sequences.keySet())
         {
            if (sequences.get(db).contains(seqName))
            {
               // I have found the database to which the sequence belongs to
               ldbName = db;
               break;
            }
         }

         // let supper class handle this:
         if (ldbName == null)
         {
            ldbName = getDatabaseForSequence(seqName);
         }

         // if still cannot find the appropriate database, return without 
         // setting anything
         if (ldbName == null)
         {
            return;
         }
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      // P2JDialect dialect = (P2JDialect) persistence.getDialect();

      try
      {
         persistence.beginTransaction();
         // step1: get sequence details:
         ResultSet rs = persistence.executeSQLQuery(
               "SELECT maxval, minval " +
               "FROM \"_P2JSequences_\" " +
               "WHERE name='" + seqName + "';");
         // H2 Dialect: "SELECT maxval, minval FROM _P2JSequences_ WHERE name='my-seq-1';"
         if (rs.next())
         {
            long seqMax = rs.getLong(1);
            long seqMin = rs.getLong(2);

            if (seqMin <= newVal && newVal <= seqMax)
            {
               // update sequence only if between bound
               persistence.executeSQL(
                     "UPDATE \"_P2JSequences_\" " +
                     "SET currval = " + newVal + ", " +
                        "unknown = false " +
                     "WHERE name='" + seqName + "';");
               // H2 Dialect: "UPDATE  _P2JSequences_ SET currval = 5, unknown = false WHERE name='my-seq-1';"
            }
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
      catch (SQLException sqle)
      {
         sqle.printStackTrace();
      }
      finally
      {
         try
         {
            persistence.commit();
         }
         catch (PersistenceException e)
         {
            e.printStackTrace();
         }
      }
   }

}
