/*
** Module   : NativeSequence.java
** Abstract : Static class that handles sequence functions and statements 
**             from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 OM  20121116          Created initial version with CURRVAL, NEXTVAL 
**                           and SETVAL support only.
*/

package com.goldencode.p2j.persist;

import java.sql.*;

import com.goldencode.p2j.persist.dialect.P2JDialect;
import com.goldencode.p2j.util.integer;

/**
 * Implementation of sequences using native SQL queries. The calls from
 * converted P2J code are delegated to protected method implementations.
 */
public class NativeSequence extends Sequence
{
   /**
    * Queries the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has already passed the maximum value.
    */
   protected integer internalCurrentValue(String seqName, String ldbName)
   {
      //Database database = ConnectionManager.get().getDatabase(ldbName);
      //P2JDialect dialect = (P2JDialect) 
      //      PersistenceFactory.getInstance(database).getDialect();

      if (ldbName == null)
      {
         ldbName = getDatabaseForSequence(seqName);
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceCurrValString(seqName);
      integer ret = new integer(); // unknown result, (TODO: promote it to int64)

      try
      {
         ResultSet rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            ret = new integer(rs.getLong(1));
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
      catch (SQLException sqle)
      {
         sqle.printStackTrace();
      }

      return ret;
   }

   /**
    * Computes and returns the next value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has passed the maximum/minimum value.
    */
   protected integer internalNextValue(String seqName, String ldbName)
   {
      if (ldbName == null)
      {
         ldbName = getDatabaseForSequence(seqName);
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceNextValString(seqName);
      integer ret = new integer(); // unknown result, (TODO: promote it to int64)

      try
      {
         ResultSet rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            ret = new integer(rs.getLong(1));
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
      catch (SQLException sqle)
      {
         sqle.printStackTrace();
      }

      return ret;
   }

   /**
    * Initialize (reset) the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   protected void internalSetValue(String seqName,
                                    String ldbName,
                                    long newVal)
   {
      if (ldbName == null)
      {
         ldbName = getDatabaseForSequence(seqName);
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceSetValString(seqName, newVal);

      try
      {
         persistence.executeSQLQuery(strQuery);
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
   }

}