/*
** Module   : TemporarySequence.java
** Abstract : Static class that handles sequence functions and statements 
**             from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 OM  20121116          Created initial version
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.integer;

/**
 * In-memory implementation of 4GL sequences. As they are not persistent,
 * their value will not be saved across sessions.
 */
public class TemporarySequence extends Sequence
{
   /**
    * Queries the current value of a sequence.
    * @param   seqName
    *          The name of the sequence
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has already passed the maximum value,
    */
   protected integer internalCurrentValue(String seqName, String ldbName)
   {
      return isUnknown ? new integer() : new integer(crtValue);
   }

   /**
    * Computes and returns the next value of a sequence.
    * @param   seqName
    *          The name of the sequence
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has passed the maximum/minimum value,
    */
   protected integer internalNextValue(String seqName, String ldbName)
   {
      // handle exceptions first
      if (isUnknown || step == 0)
      {
         return new integer();
      }

      if (step < 0 && limit > crtValue)
      {
         isUnknown = true;
         return new integer();
      }

      if (step > 0 && limit < crtValue)
      {
         isUnknown = true;
         return new integer();
      }

      crtValue += step;
      // check overflow:
      if ((step > 0 && crtValue > limit) || (step > 0 && crtValue > limit))
      {
         if (cycling)
         {
            crtValue = initialValue;
         }
         else
         {
            isUnknown = true;
            return new integer();
         }
      }

      return internalCurrentValue(seqName, ldbName);
   }

   /**
    * Initialize (reset) the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   protected void internalSetValue(String seqName,
                                    String ldbName,
                                    long newVal)
   {
      crtValue = newVal;
      isUnknown = false;
   }

   /** The initial value of the sequence */
   private long initialValue = 0;

   /** The increment. Can be positive or negative. 0 is illegal */
   private long step = 1;

   /** Cycling property. A cycling sequence will warp around to initial */
   private boolean cycling = false;

   /** Current value of the sequence. */
   private long crtValue = 0;

   /** The maximum/minimum value of the sequence. If current value passes over */
   private long limit = 1000;

   /** Flag for non-cycling sequences that passed the limit */
   private boolean isUnknown = false;

   /** Logical database name */
   private String ldbName = null;

   /** Sequence's name */
   private String name = null;
}
